;; Module for creating standard "login" and "logout" interfaces
;; Assumtions:
;; "login" - field with login name, this name may be fixed via login function parameter
;; "password" - field with login password
;; "locale" - field to change login language
;; Main functions:
;; login - create login interface, optional argument - commit callback, default value is a commit-login
;; commit-login - standard login commit callback, optional argument - login name (e.g. root), default-value (form-value "login")
;; logout -create logout interface

;; example 1: (define init login) ;; simplest case
;; example 2; (define (init) (login (lambda() (commit-login "root"))))

(define-module (alterator login)
  :use-module (alterator ahttpd) ;; uri-prefix?
  :use-module (alterator ajax)
  :use-module (alterator str) ;; string-cut
  :use-module (alterator woo)
  :use-module (alterator session)
  :use-module (alterator menu)
  :export (
           login
           commit-login
           logout))

;;; private

(define *life-time* (* 60 60 3))

(define (is-csrf-token? str)
  (and (= 13 (string-length str))
       (let ((prefix (string-append "/" str)))
         (null? (filter (lambda (item)
                          (uri-prefix? prefix
                                       (menu-item-ref item 'uri)))
                      (menu-list-item))))))

(define (update-csrf-token path token)
  (let ((path (string-split path #\/)))
    (or (and (pair? path)
             (equal? "" (car path))
             (pair? (cdr path))
             (is-csrf-token? (cadr path))
             (string-join (append (or (and token
                                           (list "" token))
                                      (list ""))
                                  (cddr path))
                          "/"))
        (string-append "/" token (string-join path "/")))))

(define (continue-path . token)
  (let ((token (and (pair? token) (car token)))
        (path (form-value "continue")))
    (or (and (string? path)
             (not (string-null? path))
             (char=? (string-ref path 0) #\/)
             (not (uri-prefix? (ahttpd-session-ref 'ahttpd-uri) path))
             (update-csrf-token path token))
        "/")))

(define (change-locale)
  (form-update-cookie "language" (form-value "locale"))
  (form-replace (ahttpd-session-ref 'ahttpd-uri) 'continue (or (form-value "continue") "/")))

;;; (userchain) reads the "loginchain" setting from the current
;;; configuration and caches it as 'ahttpd-loginchain value of
;;; the current session.
;;;
;;; The "loginchain" is configured as follows:
;;;
;;;     loginchain	user1:user2,user2:user3,...
;;;
;;; Each configured pair of usernames B:A requires that before
;;; login as the user B is allowed, a login under user A is required
;;; (see (check-next-login username) below).
;;;
;;; The procedure returns an association list, i.e.:
;;;     (("user1" . "user2") ("user2" . "user3") ...)
(define (userchain)
  (or (ahttpd-session-ref 'ahttpd-loginchain)
      (let ((*userchain* (map (lambda (s)
                                (let ((req (string-split s #\:)))
                                  (cons (car req)
                                        (if (null? (cdr req))
                                            "*"
                                            (cadr req)))))
                              (let ((userchain-str
                                     (ahttpd-config-ref "loginchain" "")))
                                (if (string-null? userchain-str)
                                  '()
                                  (string-split userchain-str #\,))))))
        (ahttpd-session-set! 'ahttpd-loginchain *userchain*)
        *userchain*)))

;;; (ahttpd-session-loginchain) returns the 'session-loginchain
;;; value of the current ahttpd session. Instead of #f the proc
;;; returns the empty list.
(define (ahttpd-session-loginchain)
  (or (ahttpd-session-ref 'session-loginchain '())
      '()))

;;; (check-next-login username) checks if a configured requirement
;;; for a login under the given user name is already fulfilled, if any.
;;; The information about pre-logins is taken from the session
;;; value 'session-loginchain. See (userchain) above and
;;; (prepend-loginchain! username) below for more information.
(define (check-next-login username)
  (let* ((*userchain* (userchain))
         (*loginchain* (ahttpd-session-loginchain))
         (*req* (assoc username *userchain*)))
    (or (not *req*)
        (equal? (cdr *req*) "*")
        (and (not (null? *loginchain*))
             (equal? (cdr *req*)
                     (car *loginchain*))))))

;;; (prepend-loginchain! username) prepends the current session list
;;; value 'session-loginchain with the given username and saves the
;;; resulting list in the given session under the same key. That way,
;;; the 'session-loginchain list grows from session to session.
;;;
;;; If the current session user name is equal to the given one
;;; no modification is made. However the list value is still copied
;;; to the new (given) session.
;;;
;;; The procedure returns the modified list.
(define (prepend-loginchain! session username)
  (let* ((*loginchain* (ahttpd-session-loginchain))
         (*loginchain* (if (not (equal? (ahttpd-session-user)
                                        username))
                        (cons username *loginchain*)
                        *loginchain*)))
    (session-set! session 'session-loginchain *loginchain*)
    *loginchain*))


;;; (prelogin-only username) checks if the given username is
;;; configured as the prelogin-only username. The configuration
;;; syntax is as follows:
;;;
;;;     prelogin-only	user1,user2,...
(define (prelogin-only username)
  (member username
          (string-split (ahttpd-config-ref "prelogin-only" "")
                        #\,)))

;;; (find-next-username) tries to find the most probable username
;;; for the next login using the "loginchain" configuration parameter
;;; and the current session login chain value 'session-loginchain.
(define (find-next-username)
  (let* ((reversechain (map (lambda (p) (cons (cdr p) (car p)))
                            (userchain)))
         (*loginchain* (ahttpd-session-loginchain))
         (lastlogin (and (not (null? *loginchain*))
                         (car *loginchain*))))
    (if lastlogin
      (let ((nextname (assoc lastlogin reversechain)))
        (or (and nextname (cdr nextname))
            ""))
      (or (and (not (null? reversechain))
               (pair? (car reversechain))
               (caar reversechain))
          "root"))))

;;; the local translator
(define (__ str)
  (_ str "alterator-ahttpd"))

;;; public interface

(define (commit-login . name)
  (let ((username (if (pair? name) (car name) (form-value "login"))))
    (catch/message
     (lambda()
       (or (check-next-login username)
           (throw 'woo-error ;; A small pretense
                  (format #f (__ "Login denied for user ~a!") username)))
       (woo-call "/login/authenticate"
                 'login username
                 'password (form-value "password")
                 'ip (message-header (ahttpd-session-ref 'ahttpd-request)
                                     "remote-addr")
                 'language (form-value "language"))
       (let ((session (make-session username *life-time*)))
         (prepend-loginchain! session username)
         (form-update-cookie "session" session)
         (display (+ (current-time) *life-time*))
         (if (prelogin-only username)
           (begin
             (newline)
             (continue-with-new-session! session relogin))
           (begin
             (woo-new "/csrf-token/new" 'session_id session 'expire (number->string (+ (current-time) *life-time*)))
             (let ((token (woo-get-option
                           (woo-read-first '/csrf-token/get 'session_id session)
                           'token
                           #f)))
               (display "goto: ")
               (display token)
               (newline)
               (form-replace (continue-path token))))))))))


(define (relogin)
  (format #t "Current userchain: ~a~%" (ahttpd-session-loginchain))
  (catch/message
   (lambda()
     (form-update-value "welcome" (__ (if (null? (ahttpd-session-loginchain))
                                        "Please login"
                                        "Please, continue with login")))
     (let ((nextname (find-next-username)))
       (form-update-value "login" nextname)
       (form-update-value "password" "")
       (if (and nextname (not (string-null? nextname)))
         (form-focus "password")
         (form-focus "login")))
     (form-update-enum "locale" (woo-call "/login/list_locale"))
     (form-update-value "locale" (car (string-cut (form-value "language") #\;))))))

(define (login . commit)
  (relogin)
  (let ((commit (if (pair? commit) (car commit) commit-login)))
    (form-bind "locale" "change" change-locale)
    (form-bind "commit" "click" commit)
    (form-bind "password" "enter" commit)))

(define (logout)
  (woo-delete '/csrf-token/delete 'session_id (form-cookie "session" ))
  (session-delete (form-cookie "session"))
  (form-update-cookie "session" "")
  (form-replace "/"))
